import threading
from typing import Callable, Any
import json
import time
from pynostr.event import Event
from nostr_agents.nostr_client import NostrClient
from mcp.server.fastmcp.exceptions import ToolError
from mcp.server.fastmcp.tools.tool_manager import ToolManager


class NostrMCPServer(object):
    def __init__(self, display_name: str, nostr_client: NostrClient):
        self.display_name = display_name
        self.client = nostr_client
        self.tool_to_sats_map = {}
        self.tool_manager = ToolManager()

    def add_tool(self,
                 fn: Callable[..., Any],
                 name: str | None = None,
                 description: str | None = None,
                 satoshis: int | None = None):
        if satoshis:
            self.tool_to_sats_map[name or fn.__name__] = satoshis
        self.tool_manager.add_tool(
            fn=fn,
            name=name,
            description=description,
        )

    def list_tools(self) -> dict[str, Any]:
        """Define available tools"""
        return {
            "tools": [{
                "name": tool.name,
                "description": tool.description,
                "inputSchema": tool.parameters,
                "satoshis": self.tool_to_sats_map.get(tool.name, 0),
            } for tool in self.tool_manager.list_tools()]
        }

    def call_tool(
        self,
        name: str,
        arguments: dict[str, Any],
    ) -> Any:
        """Call a tool by name with arguments."""
        tool = self.tool_manager.get_tool(name)
        if not tool:
            raise ToolError(f"Unknown tool: {name}")
        result = tool.fn(**arguments)
        return result

    def _direct_message_callback(self, event: Event, message: str):
        """
        Callback function to handle incoming direct messages.
        :param event: The event object containing the message.
        :param message: The message content.
        """
        # Process the incoming message
        message = message.strip()
        print(f"Request: {message}")
        try:
            request = json.loads(message)
            if request['action'] == 'list_tools':
                response = self.list_tools()

            elif request['action'] == 'call_tool':
                tool_name = request['tool_name']
                arguments = request['arguments']

                satoshis = self.tool_to_sats_map.get(tool_name, 0)
                if satoshis > 0:
                    # Requires payment first
                    invoice = self.client.nwc_client.make_invoice(amt=satoshis,
                                                                  desc="Payment for tool call")
                    response = invoice

                    def on_success():
                        print(f"Payment succeeded for {tool_name}")
                        result = self.call_tool(tool_name, arguments)
                        response = {
                            "content": [{
                                "type": "text",
                                "text": str(result)
                            }]
                        }
                        print(f'On success response: {response}')
                        thr = threading.Thread(
                            target=self.client.send_direct_message_to_pubkey,
                            args=(event.pubkey, json.dumps(response)),
                        )
                        thr.start()

                    def on_failure():
                        response = {
                            "error": f"Payment failed for {tool_name}"
                        }
                        print(f"On failure response: {response}")
                        thr = threading.Thread(
                            target=self.client.send_direct_message_to_pubkey,
                            args=(event.pubkey, json.dumps(response)),
                        )
                        thr.start()
                    thr = threading.Thread(
                        target=self.client.nwc_client.on_payment_success,
                        kwargs={
                            'invoice': invoice,
                            'callback': on_success,
                            'timeout': 20,
                            'unsuccess_callback': on_failure,
                        }
                    )
                    thr.start()
                else:
                    result = self.call_tool(tool_name, arguments)
                    response = {
                        "content": [{
                            "type": "text",
                            "text": str(result)
                        }]
                    }

            else:
                response = {
                    "error": f"Invalid action: {request['action']}"
                }
        except Exception as e:
            response = {
                "content": [{
                    "type": "text",
                    "text": str(e)
                }]
            }

        if not isinstance(response, str):
            response = json.dumps(response)

        print(f'Response: {response}')
        time.sleep(1)
        thr = threading.Thread(
            target=self.client.send_direct_message_to_pubkey,
            args=(event.pubkey, response),
        )
        thr.start()

    def start(self):
        # Update list_tools metadata
        thr = threading.Thread(
            target=self.client.update_metadata,
            kwargs={
                'name': 'mcp_server',
                'display_name': self.display_name,
                'about': json.dumps(self.list_tools())
            }
        )
        print(f'Updating metadata for {self.client.public_key.bech32()}')
        thr.start()
        time.sleep(5)

        print(f'Starting message listener for {self.client.public_key.bech32()}')

        # Start call_tool listener
        self.client.direct_message_listener(
            callback=self._direct_message_callback
        )


if __name__ == "__main__":
    import os
    from dotenv import load_dotenv

    load_dotenv()

    # Get the environment variables
    relays = os.getenv('NOSTR_RELAYS').split(',')
    private_key = os.getenv('NOSTR_SERVER_PRIVATE_KEY')
    nwc_str = os.getenv('NWC_CONN_STR')


    def add(a: int, b: int) -> int:
        """Add two numbers"""
        return a + b

    def multiply(a: int, b: int) -> int:
        """Multiply two numbers"""
        return a * b

    def get_weather(city: str) -> str:
        """Gets the weather for a city"""
        if city.lower() in {'portland', 'seattle', 'vancouver'}:
            return 'rainy'
        else:
            return 'sunny'

    def get_current_date() -> str:
        """Gets today's date"""
        return time.strftime("%Y-%m-%d")

    def get_current_time() -> str:
        """Gets the time of day"""
        return time.strftime("%H:%M:%S")

    # Create an instance of NostrClient
    client = NostrClient(relays, private_key, nwc_str)
    server = NostrMCPServer("Evan's MCP Server", client)
    server.add_tool(add)  # Add by signature alone
    server.add_tool(multiply, name="multiply", description="Multiply two numbers")  # Add by signature and name
    server.add_tool(get_weather, satoshis=10)  # Specify price in satoshis
    server.add_tool(get_current_date)
    server.add_tool(get_current_time, satoshis=5)

    server.start()

    '''
    {"action": "list_tools"}
    {"action": "call_tool", "tool_name": "add", "arguments": {"a": 1, "b": 2}}
    {"action": "call_tool", "tool_name": "multiply", "arguments": {"a": 2, "b": 5}}
    '''



from typing import List, Any, Optional, Callable
import logging
import uuid
import time

from pynostr.base_relay import RelayPolicy
from pynostr.key import PrivateKey
from pynostr.message_type import RelayMessageType
from pynostr.relay_manager import RelayManager
from pynostr.event import EventKind
from pynostr.filters import Filters, FiltersList
from pynostr.encrypted_dm import EncryptedDirectMessage
from pynostr.metadata import Metadata
from pynostr.event import Event
from pynostr.utils import get_public_key, get_timestamp

from nostr_agents.nwc_client import NWCClient


logging.basicConfig(level=logging.WARNING)  # Set the minimum logging level


logger = logging.getLogger(__name__)


def log_callback(*args):
    logging.info(f"Received message from {args}")


class NostrClient(object):
    def __init__(self,
                 relays: List[str],
                 private_key: str,
                 nwc_str: str = None,
                 ):
        """
        Initialize the NostrClient with a list of relays, private key, and NWC string.
        :param relays: List of relay URLs.
        :param private_key: nsec private key.
        :param nwc_str: Nostr Wallet Connection string (optional).
        """
        self.relays = relays
        self.private_key = PrivateKey.from_nsec(private_key)
        self.public_key = self.private_key.public_key
        self.nwc_client = NWCClient(nwc_str) if nwc_str else None

    def sign(self, event: Event):
        """Sign the event with the private key."""
        event.sign(self.private_key.hex())
        return event

    def get_relay_manager(self,
                          message_callback=log_callback,
                          timeout: int = 5,
                          error_threshold: int = 3,
                          close_on_eose: bool = False,
                          policy: RelayPolicy = RelayPolicy()) -> RelayManager:
        relay_manager = RelayManager(timeout=timeout,
                                     error_threshold=error_threshold)

        for relay in self.relays:
            relay_manager.add_relay(relay.strip(),
                                    close_on_eose=close_on_eose,
                                    policy=policy,
                                    timeout=timeout,
                                    message_callback=message_callback)
        return relay_manager

    def get_metadata_for_pubkey(self, public_key: str | PrivateKey = None) -> Optional[Metadata]:
        relay_manager = self.get_relay_manager()
        public_key = get_public_key(public_key if isinstance(public_key, str) else public_key.hex()) if public_key else self.public_key
        filters = FiltersList(
            [  # enter filter condition
                Filters(
                    kinds=[EventKind.SET_METADATA],
                    authors=[
                        public_key.hex(),
                    ],
                    limit=1
                )
            ]
        )

        subscription_id = uuid.uuid1().hex
        relay_manager.add_subscription_on_all_relays(subscription_id, filters)
        relay_manager.run_sync()
        messages = []
        while relay_manager.message_pool.has_ok_notices():
            ok_msg = relay_manager.message_pool.get_ok_notice()
            logger.info(ok_msg)
        while relay_manager.message_pool.has_events():
            event_msg = relay_manager.message_pool.get_event()
            logger.info(event_msg.event.to_dict())
            messages.append(event_msg.event.to_dict())
        relay_manager.close_subscription_on_all_relays(subscription_id)

        if len(messages) > 0:
            latest_metadata: dict = sorted(messages, key=lambda x: x['created_at'], reverse=True)[0]
            return Metadata.from_dict(latest_metadata)
        else:
            return None

    def update_metadata(self,
                        name: Optional[str] = None,
                        about: Optional[str] = None,
                        nip05: Optional[str] = None,
                        picture: Optional[str] = None,
                        banner: Optional[str] = None,
                        lud16: Optional[str] = None,
                        lud06: Optional[str] = None,
                        username: Optional[str] = None,
                        display_name: Optional[str] = None,
                        website: Optional[str] = None):
        previous_metadata = self.get_metadata_for_pubkey(self.public_key)

        metadata = Metadata()
        if previous_metadata:
            metadata.set_metadata(previous_metadata.metadata_to_dict())

        if name:
            metadata.name = name
        if about:
            metadata.about = about
        if nip05:
            metadata.nip05 = nip05
        if picture:
            metadata.picture = picture
        if banner:
            metadata.banner = banner
        if lud16:
            metadata.lud16 = lud16
        if lud06:
            metadata.lud06 = lud06
        if username:
            metadata.username = username
        if display_name:
            metadata.display_name = display_name
        if website:
            metadata.website = website

        metadata.created_at = int(time.time())
        metadata.update()

        if previous_metadata and previous_metadata.content == metadata.content:
            print("No changes in metadata, skipping update.")
            return

        event = self.sign(metadata.to_event())
        relay_manager = self.get_relay_manager(timeout=10)

        relay_manager.publish_event(event)
        relay_manager.run_sync()

    def send_direct_message_to_pubkey(self,
                                      recipient_pubkey: str,
                                      message: str):
        recipient = get_public_key(recipient_pubkey)
        dm = EncryptedDirectMessage()
        dm.encrypt(
            self.private_key.hex(),
            cleartext_content=message,
            recipient_pubkey=recipient.hex(),
        )

        dm_event = dm.to_event()
        dm_event.sign(self.private_key.hex())

        relay_manager = self.get_relay_manager()
        relay_manager.publish_message(dm_event.to_message())
        relay_manager.run_sync()

    def direct_message_listener(self,
                                callback: Callable[[Event, str], Any],
                                recipient_pubkey: str = None,
                                timeout: int = 0,
                                close_after_first_message: bool = False):
        if recipient_pubkey:
            authors = [get_public_key(recipient_pubkey).hex()]
        else:
            authors = None

        filters = FiltersList(
            [
                Filters(
                    authors=authors,
                    kinds=[EventKind.ENCRYPTED_DIRECT_MESSAGE],
                    since=get_timestamp(),
                    limit=10,
                )
            ]
        )

        subscription_id = uuid.uuid1().hex

        ack = set([])

        def print_dm(message_json, *args):
            message_type = message_json[0]
            success = False
            if message_type == RelayMessageType.EVENT:
                event = Event.from_dict(message_json[2])
                if event.kind == EventKind.ENCRYPTED_DIRECT_MESSAGE:
                    if event.id in ack:
                        return
                    ack.add(event.id)
                    if event.has_pubkey_ref(self.public_key.hex()):
                        rdm = EncryptedDirectMessage.from_event(event)
                        rdm.decrypt(self.private_key.hex(), public_key_hex=event.pubkey)
                        success = callback(event, rdm.cleartext_content)
                        logging.info(f"New dm received:{event.date_time()} {rdm.cleartext_content}")
            elif message_type == RelayMessageType.OK:
                logging.info(message_json)
            elif message_type == RelayMessageType.NOTICE:
                logging.info(message_json)
            if success and close_after_first_message:
                relay_manager.close_subscription_on_all_relays(subscription_id)

        relay_manager = self.get_relay_manager(message_callback=print_dm, timeout=timeout)
        relay_manager.add_subscription_on_all_relays(subscription_id, filters)
        relay_manager.run_sync()



from websocket import create_connection
import json
import base64
import time
import math
import hashlib
from secp256k1 import PrivateKey, PublicKey
from Crypto import Random
from Crypto.Cipher import AES
import threading

BS = 16
pad = lambda s: s + (BS - len(s) % BS) * chr(BS - len(s) % BS)
unpad = lambda s: s[:-ord(s[len(s) - 1:])]


def encrypt(privkey, pubkey, plaintext):
    key = PublicKey(bytes.fromhex("02" + pubkey), True).tweak_mul(bytes.fromhex(privkey)).serialize().hex()[2:]
    key_bytes = 32
    key = bytes.fromhex(key)
    plaintext = pad(plaintext)
    plaintext = plaintext.encode("utf-8")
    assert len(key) == key_bytes

    # Choose a random, 16-byte IV.
    iv = Random.new().read(AES.block_size)

    # Create AES-CTR cipher.
    aes = AES.new(key, AES.MODE_CBC, iv)

    # Encrypt and return IV and ciphertext.
    ciphertext = aes.encrypt(plaintext)

    # Convert to base64
    cipher_b64 = base64.b64encode(ciphertext).decode('ascii')
    cipher_iv = base64.b64encode(iv).decode('ascii')

    return cipher_b64 + "?iv=" + cipher_iv


def decrypt(privkey, pubkey, ciphertext):
    key = PublicKey(bytes.fromhex("02" + pubkey), True).tweak_mul(bytes.fromhex(privkey)).serialize().hex()[2:]
    key_bytes = 32
    key = bytes.fromhex(key)
    (ciphertext, iv) = ciphertext.split("?iv=")
    ciphertext = base64.b64decode(ciphertext.encode('ascii'))
    iv = base64.b64decode(iv.encode('ascii'))
    assert len(key) == key_bytes

    # Create AES-CTR cipher.
    aes = AES.new(key, AES.MODE_CBC, iv)

    # Decrypt and return the plaintext.
    plaintext = aes.decrypt(ciphertext).decode('ascii')
    plaintext = unpad(plaintext)
    return plaintext


def sha256(text_to_hash):
    m = hashlib.sha256()
    m.update(bytes(text_to_hash, 'UTF-8'))
    return m.digest().hex()


def processNWCstring(string):
    if (string[0:22] != "nostr+walletconnect://"):
        print('Your pairing string was invalid, try one that starts with this: nostr+walletconnect://')
        return
    string = string[22:]
    arr = string.split("&")
    item = arr[0].split("?")
    del arr[0]
    arr.insert(0, item[0])
    arr.insert(1, item[1])
    arr[0] = "wallet_pubkey=" + arr[0]
    arr2 = []
    obj = {}
    for item in arr:
        item = item.split("=")
        arr2.append(item[0])
        arr2.append(item[1])
    for index, item in enumerate(arr2):
        if (item == "secret"):
            arr2[index] = "app_privkey"
    for index, item in enumerate(arr2):
        if (index % 2):
            obj[arr2[index - 1]] = item
    obj["app_pubkey"] = PrivateKey(bytes.fromhex(obj["app_privkey"])).pubkey.serialize().hex()[2:]
    return obj


def getEvents(relay, ids, kinds, until, since, limit, etags, ptags):
    events = []
    subId = PrivateKey().serialize()[0:16]
    myfilter = {}
    if (ids):
        myfilter["ids"] = ids
    if (kinds):
        myfilter["kinds"] = kinds
    if (until):
        myfilter["until"] = until
    if (since):
        myfilter["since"] = since
    if (limit):
        myfilter["limit"] = limit
    if (etags):
        myfilter["#e"] = etags
    if (ptags):
        myfilter["#p"] = ptags
    subscription = ["REQ", subId, myfilter]
    ws = create_connection(relay)
    ws.send(json.dumps(subscription))
    for i in range(limit + 1):
        response = ws.recv()
        response = json.loads(response)
        if (len(response) < 3):
            continue
        events.append(response[2])
        ws.close()
    return events


def getResponse(nwc_obj, event_id, val):
    relay = nwc_obj["relay"]
    ids = None
    kinds = [23195]
    until = None
    since = None
    limit = 1
    etags = [event_id]
    ptags = [nwc_obj["app_pubkey"]]
    events = []
    for i in [1, 2, 3, 4, 5, 6, 7]:
        if (not len(events)):
            events2 = getEvents(relay, ids, kinds, until, since, limit, etags, ptags)
            if (not not len(events2)): events = events2
    if (not not len(events)):
        val[0] = events[0]
        return
    val[0] = events
    return


def sendEvent(event, nwc_obj):
    event_id = json.loads(event)[1]["id"]
    relay = nwc_obj["relay"]
    response = None
    ws = create_connection(relay)
    ws.send(event)
    response = ws.recv()
    ws.close()
    return response


def getSignedEvent(event, privkey):
    eventData = json.dumps([
        0,
        event['pubkey'],
        event['created_at'],
        event['kind'],
        event['tags'],
        event['content']
    ], separators=(',', ':'))
    event["id"] = sha256(eventData);
    privkey = PrivateKey(bytes.fromhex(privkey))
    event["sig"] = privkey.schnorr_sign(bytes.fromhex(event["id"]), "none", True).hex()
    return event


def makeInvoice(nwc_obj, amt, desc):
    msg = json.dumps({
        "method": "make_invoice",
        "params": {
            "amount": amt * 1000,
            "description": desc,
        } if amt else {
            "description": desc,
        }
    })
    emsg = encrypt(nwc_obj["app_privkey"], nwc_obj["wallet_pubkey"], msg);
    obj = {
        "kind": 23194,
        "content": emsg,
        "tags": [["p", nwc_obj["wallet_pubkey"]]],
        "created_at": math.floor(time.time()),
        "pubkey": nwc_obj["app_pubkey"],
    }
    event = getSignedEvent(obj, nwc_obj["app_privkey"])
    eid = event["id"]
    event = json.dumps(["EVENT", event], separators=(',', ':'))
    val = [False]
    download_thread = threading.Thread(target=getResponse, name="Background", args=(nwc_obj, eid, val))
    download_thread.start()
    sendEvent(event, nwc_obj)
    for i in [1, 2, 3]:
        if (not val[0]):
            time.sleep(1)
            continue
    response = val[0]
    ersp = response["content"]
    drsp = decrypt(nwc_obj["app_privkey"], nwc_obj["wallet_pubkey"], ersp)
    dobj = json.loads(drsp)
    return dobj


def checkInvoice(nwc_obj, invoice=None, payment_hash=None):
    if invoice is None and payment_hash is None:
        raise ValueError("Either 'invoice' or 'payment_hash' must be provided")

    params = {}
    if invoice is not None:
        params["invoice"] = invoice
    if payment_hash is not None:
        params["payment_hash"] = payment_hash

    msg = json.dumps({
        "method": "lookup_invoice",
        "params": params
    })
    emsg = encrypt(nwc_obj["app_privkey"], nwc_obj["wallet_pubkey"], msg);
    obj = {
        "kind": 23194,
        "content": emsg,
        "tags": [["p", nwc_obj["wallet_pubkey"]]],
        "created_at": math.floor(time.time()),
        "pubkey": nwc_obj["app_pubkey"],
    }
    event = getSignedEvent(obj, nwc_obj["app_privkey"])
    eid = event["id"]
    event = json.dumps(["EVENT", event], separators=(',', ':'))
    val = [False]
    download_thread = threading.Thread(target=getResponse, name="Background", args=(nwc_obj, eid, val))
    download_thread.start()
    sendEvent(event, nwc_obj)
    for i in [1, 2, 3]:
        if (not val[0]):
            time.sleep(1)
            continue
    response = val[0]
    ersp = response["content"]
    drsp = decrypt(nwc_obj["app_privkey"], nwc_obj["wallet_pubkey"], ersp)
    dobj = json.loads(drsp)
    return dobj
    # an error looks like this:
    # {error: {code: "INTERNAL", message: "Something went wrong while looking up invoice: "}, result_type: "lookup_invoice"}


def didPaymentSucceed(nwc_obj, invoice):
    invoice_info = checkInvoice(nwc_obj, invoice=invoice)
    if (invoice_info and not ("error" in invoice_info) and ("result" in invoice_info) and (
            "preimage" in invoice_info["result"])):
        return invoice_info["result"]["preimage"]
    return False


def tryToPayInvoice(nwc_obj, invoice, amnt=None):
    msg = {
        "method": "pay_invoice",
        "params": {
            "invoice": invoice,
        }
    }
    if (amnt): msg["params"]["amount"] = amnt
    msg = json.dumps(msg)
    emsg = encrypt(nwc_obj["app_privkey"], nwc_obj["wallet_pubkey"], msg);
    obj = {
        "kind": 23194,
        "content": emsg,
        "tags": [["p", nwc_obj["wallet_pubkey"]]],
        "created_at": math.floor(time.time()),
        "pubkey": nwc_obj["app_pubkey"],
    }
    event = getSignedEvent(obj, nwc_obj["app_privkey"])
    eid = event["id"]
    event = json.dumps(["EVENT", event], separators=(',', ':'))
    sendEvent(event, nwc_obj)


def getInfo(nwc_obj):
    msg = {
        "method": "get_info"
    }
    msg = json.dumps(msg)
    emsg = encrypt(nwc_obj["app_privkey"], nwc_obj["wallet_pubkey"], msg);
    obj = {
        "kind": 23194,
        "content": emsg,
        "tags": [["p", nwc_obj["wallet_pubkey"]]],
        "created_at": math.floor(time.time()),
        "pubkey": nwc_obj["app_pubkey"],
    }
    event = getSignedEvent(obj, nwc_obj["app_privkey"])
    eid = event["id"]
    event = json.dumps(["EVENT", event], separators=(',', ':'))
    val = [False]
    download_thread = threading.Thread(target=getResponse, name="Background", args=(nwc_obj, eid, val))
    download_thread.start()
    sendEvent(event, nwc_obj)
    for i in [1, 2, 3]:
        if (not val[0]):
            time.sleep(1)
            continue
    response = val[0]
    ersp = response["content"]
    drsp = decrypt(nwc_obj["app_privkey"], nwc_obj["wallet_pubkey"], ersp)
    dobj = json.loads(drsp)
    return dobj


def listTx(nwc_obj, params={}):
    msg = {
        "method": "list_transactions",
        "params": params
    }
    msg = json.dumps(msg)
    emsg = encrypt(nwc_obj["app_privkey"], nwc_obj["wallet_pubkey"], msg);
    obj = {
        "kind": 23194,
        "content": emsg,
        "tags": [["p", nwc_obj["wallet_pubkey"]]],
        "created_at": math.floor(time.time()),
        "pubkey": nwc_obj["app_pubkey"],
    }
    event = getSignedEvent(obj, nwc_obj["app_privkey"])
    eid = event["id"]
    event = json.dumps(["EVENT", event], separators=(',', ':'))
    val = [False]
    download_thread = threading.Thread(target=getResponse, name="Background", args=(nwc_obj, eid, val))
    download_thread.start()
    sendEvent(event, nwc_obj)
    for i in [1, 2, 3]:
        if (not val[0]):
            time.sleep(1)
            continue
    response = val[0]
    ersp = response["content"]
    drsp = decrypt(nwc_obj["app_privkey"], nwc_obj["wallet_pubkey"], ersp)
    dobj = json.loads(drsp)
    return dobj


def getBalance(nwc_obj):
    msg = {
        "method": "get_balance"
    }
    msg = json.dumps(msg)
    emsg = encrypt(nwc_obj["app_privkey"], nwc_obj["wallet_pubkey"], msg);
    obj = {
        "kind": 23194,
        "content": emsg,
        "tags": [["p", nwc_obj["wallet_pubkey"]]],
        "created_at": math.floor(time.time()),
        "pubkey": nwc_obj["app_pubkey"],
    }
    event = getSignedEvent(obj, nwc_obj["app_privkey"])
    eid = event["id"]
    event = json.dumps(["EVENT", event], separators=(',', ':'))
    val = [False]
    download_thread = threading.Thread(target=getResponse, name="Background", args=(nwc_obj, eid, val))
    download_thread.start()
    sendEvent(event, nwc_obj)
    for i in [1, 2, 3]:
        if (not val[0]):
            time.sleep(1)
            continue
    response = val[0]
    ersp = response["content"]
    drsp = decrypt(nwc_obj["app_privkey"], nwc_obj["wallet_pubkey"], ersp)
    dobj = json.loads(drsp)
    return dobj

# print( makeInvoice( processNWCstring( nwc_string ), 100, "test" ) )
# print( checkInvoice( processNWCstring( nwc_string ), invoice="lnbc1u1png6lw0pp5h4l73ajf4u548ktalztfwt7k9wtp9xhgqs6t0my0mw450nfkmnrsdq8w3jhxaqcqzzsxqyz5vqsp5asqxxjr2uhsfxyjwt2gxrq38dejkr76rmzl0zstjqlx8rrlcqpns9qxpqysgq22dwgadd7xnsnn8jzwkfxwy7nwclzt4d8wa3adrml83a0nvgy2hzm565k4qn0rcrzx7n2j8dszq9yqvhdx2z0xes77j5e480clx6d7cq2vvqj5" ) )
# print( didPaymentSucceed( processNWCstring( nwc_string ), "lnbc700n1pngmqvkpp57yg7u02n2pxack552mwdl5k8derwsyrgh2uft0lptqvcw8qv9l0qdpuge6kuerfdenjqsrnw4cx2un5v4ehgmn9wssx7m3qwd6xzcmtv4ezumn9waescqzzsxqrrsssp5uy70kfvlwfw4xhlu0k7hr7luq0qwgl5sdc9lyk4aqxvqzqqesjes9qyyssq9w7dyt6e64dyhws70qkvnauq59vmkh9lt4j5t598x3f7xzzv5edyg2g0rtdphtqmkqq3xja27kz4gvdgdy7qeymtms32d82gpmtekvspeyp4rq" ) )
# print( tryToPayInvoice( processNWCstring( nwc_string ), "lnbc700n1pngmqvkpp57yg7u02n2pxack552mwdl5k8derwsyrgh2uft0lptqvcw8qv9l0qdpuge6kuerfdenjqsrnw4cx2un5v4ehgmn9wssx7m3qwd6xzcmtv4ezumn9waescqzzsxqrrsssp5uy70kfvlwfw4xhlu0k7hr7luq0qwgl5sdc9lyk4aqxvqzqqesjes9qyyssq9w7dyt6e64dyhws70qkvnauq59vmkh9lt4j5t598x3f7xzzv5edyg2g0rtdphtqmkqq3xja27kz4gvdgdy7qeymtms32d82gpmtekvspeyp4rq" ) )
# print( listTx( processNWCstring( nwc_string, { "type": "outgoing"} ) ) )
# print( getBalance( processNWCstring( nwc_string ) ) )